<?php
/***************************************************************
*  Copyright notice
*
*  (c) 2013 IMIA net based solutions (d.frerich@imia.de)
*  All rights reserved
*
*  This script is part of the TYPO3 project. The TYPO3 project is
*  free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  The GNU General Public License can be found at
*  http://www.gnu.org/copyleft/gpl.html.
*
*  This script is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  This copyright notice MUST APPEAR in all copies of the script!
***************************************************************/

namespace IMIA\ImiaBaseExt\Domain\Repository;

use TYPO3\CMS\Extbase\Persistence\QueryInterface;

/**
 * @package     imia_base_ext
 * @subpackage  Domain\Repository
 * @author      David Frerich <d.frerich@imia.de>
 */
class Repository extends \TYPO3\CMS\Extbase\Persistence\Repository
{
    /**
     * @var array
     */
    protected $defaultOrderings = array(
        'sorting' => QueryInterface::ORDER_ASCENDING
    );

    /**
     * @var boolean
     *
     * @deprecated
     */
    protected $respectEnableFields = true;

    /**
     * @var boolean
     */
    protected $respectStoragePage = true;

    /**
     * @var boolean
     */
    protected $respectSysLanguage = true;

    /**
     * @var boolean
     */
    protected $ignoreEnableFields = false;

    /**
     * @var boolean
     */
    protected $includeDeleted = false;

    /**
     * @var mixed
     */
    protected $languageOverlayMode = null;

    /**
     * @var string
     */
    protected $languageMode = null;

    /**
     * @var integer
     */
    protected $languageUid = null;

    /**
     * @var array
     */
    protected $enableFieldsToBeIgnored = null;

    /**
     * @var array
     */
    protected $storagePageIds = null;

    /**
     * @param integer $limit
     * @param integer $offset
     * @param array $sorting
     * @return \TYPO3\CMS\Extbase\Persistence\QueryResultInterface
     */
	public function findAll($limit = null, $offset = null, $sorting = null)
    {
		return $this->createQuery($limit, $offset, $sorting)->execute();
	}

    /**
     * @param array $uids
     * @param integer $limit
     * @param integer $offset
     * @param array $sorting
     * @return \TYPO3\CMS\Extbase\Persistence\QueryResultInterface
     */
    public function findByUids(array $uids, $limit = null, $offset = null, $sorting = null)
    {
        if (count($uids) > 0) {
            $query = $this->createQuery($limit, $offset, $sorting);
            $constraints = array(
                $query->in('uid', $uids),
            );

            $transPointerField = $GLOBALS['TCA'][$this->getTableName()]['ctrl']['transOrigPointerField'];
            if ($transPointerField) {
                $constraints[] = $query->in($transPointerField, $uids);
            }

            $query->matching($query->logicalOr($constraints));

            return $query->execute();
        } else {
            return array();
        }
    }

    /**
     * @param array $uids
     * @return array|\TYPO3\CMS\Extbase\Persistence\QueryResultInterface
     */
    public function findByInUid(array $uids)
    {
        $query = $this->createQuery();
        $constraints = array(
            $query->in('uid', $uids),
        );

        $transPointerField = $GLOBALS['TCA'][$this->getTableName()]['ctrl']['transOrigPointerField'];
        if ($transPointerField) {
            $constraints[] = $query->in($transPointerField, $uids);
        }

        $query->matching($query->logicalOr($constraints));

        return $query->execute();
    }

    /**
     * @param integer $pid
     * @param integer $limit
     * @param integer $offset
     * @param array $sorting
     * @return \TYPO3\CMS\Extbase\Persistence\QueryResultInterface
     */
    public function findByPid($pid, $limit = null, $offset = null, $sorting = null)
    {
        $query = $this->createQuery($limit, $offset, $sorting);
        $query->matching(
            $query->equals('pid', $pid)
        );

        return $query->execute();
    }

    /**
     * @return integer
     */
    public function findLastModified($table = null)
    {
        $query = $this->createQuery(1, 0, array('tstamp', 'desc'));
        $result = $this->executeArrayQuery($query, array('tstamp'), $table);

        $lastModified = 0;
        foreach ($result as $row) {
            $lastModified = $row['tstamp'];
            break;
        }

        return $lastModified;
    }

    /**
     * @param integer $limit
     * @param integer $offset
     * @param array $sorting
     * @return QueryInterface
     */
    public function createQuery($limit = null, $offset = null, $sorting = null)
    {
        $query = parent::createQuery();

        if (!$this->respectEnableFields || $this->ignoreEnableFields) {
            $query->getQuerySettings()->setIgnoreEnableFields(true);
        }
        if (!$this->respectEnableFields || $this->includeDeleted) {
            $query->getQuerySettings()->setIncludeDeleted(true);
        }
        if (!$this->respectStoragePage) {
            $query->getQuerySettings()->setRespectStoragePage(false);
        }
        if (!$this->respectSysLanguage) {
            $query->getQuerySettings()->setRespectSysLanguage(false);
        }
        if ($this->languageOverlayMode !== null){
            $query->getQuerySettings()->setLanguageOverlayMode($this->languageOverlayMode);
        }
        if ($this->languageMode !== null){
            $query->getQuerySettings()->setLanguageMode($this->languageMode);
        }
        if ($this->languageUid !== null){
            $query->getQuerySettings()->setLanguageUid($this->languageUid);
        }
        if ($this->enableFieldsToBeIgnored !== null){
            $query->getQuerySettings()->setEnableFieldsToBeIgnored($this->enableFieldsToBeIgnored);
        }
        if ($this->storagePageIds !== null){
            $query->getQuerySettings()->setStoragePageIds($this->storagePageIds);
        }

		if ($limit && $limit > 0) {
			$query->setLimit($limit);
        }
		if ($offset && $offset > 0) {
			$query->setOffset($offset);
        }
        if ($sorting && is_array($sorting) && $sorting[0]) {
            $query->setOrderings(array(
                $sorting[0]	=> strtolower($sorting[1]) == 'asc' ?
                    QueryInterface::ORDER_ASCENDING :
                    QueryInterface::ORDER_DESCENDING,
            ));
        }

        return $query;
    }

    /**
     * @param boolean $respectEnableFields
     * @return \IMIA\ImiaBaseExt\Domain\Repository\Repository
     *
     * @deprecated
     */
    public function setRespectEnableFields($respectEnableFields)
    {
		$this->setIgnoreEnableFields(!$respectEnableFields);
		$this->setIncludeDeleted(!$respectEnableFields);

        return $this;
    }

    /**
     * @param QueryInterface $query
     * @return array
     */
    protected function executeUidArrayQuery($query, $table = null)
    {
        $uids = array();
        $result = $this->executeArrayQuery($query, array('uid'), $table);
        foreach ($result as $row) {
            $uids[] = $row['uid'];
        }

        return $uids;
    }

    /**
     * @param QueryInterface $query
     * @param array $fields
     * @return array
     */
    protected function executeArrayQuery($query, array $fields = array('uid'), $table = null)
    {
        if ((float)TYPO3_version < 6.2) {
            if (!$table) {
                $dataMapper = $this->objectManager->get('TYPO3\CMS\Extbase\Persistence\Generic\Mapper\DataMapper');
                $table = $dataMapper->convertClassNameToTableName($this->objectType);
            }

            $storageBackend = $this->objectManager->get('TYPO3\CMS\Extbase\Persistence\Generic\Storage\Typo3DbBackend');
            $parameters = array();
            $statementParts = $storageBackend->parseQuery($query, $parameters);
            $statementParts['fields'][$table] = $table . '.' . implode(',' . $table . '.', $fields);

            $statementSql = $storageBackend->buildQuery($statementParts, $parameters);

            $query = $this->createQuery();
            $query->statement($statementSql, $parameters);
        } else {
            $query->getQuerySettings()->setOnlyFields($fields);
        }

        $query->getQuerySettings()->setReturnRawQueryResult(true);

        return $query->execute();
    }

    /**
     * @param \TYPO3\CMS\Extbase\Persistence\Query $query
     * @param array $parameters
     * @return string
     */
    protected function getQuerySql($query, &$parameters = null)
    {
        if (!$parameters) {
            $parameters = array();
        }

        $storageBackend = $this->objectManager->get('TYPO3\CMS\Extbase\Persistence\Generic\Storage\Typo3DbBackend');
        $statementParts = $storageBackend->parseQuery($query, $parameters);

        return $storageBackend->buildQuery($statementParts, $parameters);
    }

    /**
     * @inheritdoc
     */
    public function __call($methodName, $arguments)
    {
        if (substr($methodName, 0, 8) === 'findByIn' && strlen($methodName) > 8 && ctype_upper(substr($methodName, 8, 1))) {
            $propertyName = lcfirst(substr($methodName, 8));
            $query = $this->createQuery();
            $query->matching($query->in($propertyName, $arguments[0]));

            return $query->execute();
        } elseif (substr($methodName, 0, 11) === 'findOneByIn' && strlen($methodName) > 11 && ctype_upper(substr($methodName, 11, 1))) {
            $propertyName = lcfirst(substr($methodName, 11));
            $query = $this->createQuery();
            $query->matching($query->in($propertyName, $arguments[0]));

            return $query->execute()->current();
        } elseif (substr($methodName, 0, 12) === 'findUidsByIn' && strlen($methodName) > 12 && ctype_upper(substr($methodName, 12, 1))) {
            $propertyName = lcfirst(substr($methodName, 12));
            $query = $this->createQuery();
            $query->matching($query->in($propertyName, $arguments[0]));

            return $this->executeUidArrayQuery($query);
        } elseif (substr($methodName, 0, 10) === 'findUidsBy' && strlen($methodName) > 10 && ctype_upper(substr($methodName, 10, 1))) {
            $propertyName = lcfirst(substr($methodName, 10));
            $query = $this->createQuery();
            $query->matching($query->in($propertyName, $arguments[0]));

            return $this->executeUidArrayQuery($query->execute());
        } else {
            return parent::__call($methodName, $arguments);
        }
    }

    public function truncate()
    {
        $GLOBALS['TYPO3_DB']->exec_TRUNCATEquery($this->getTableName());

        return true;
    }

    /**
     * @return string
     */
    public function getTableName()
    {
        return $this->objectManager
            ->get('TYPO3\CMS\Extbase\Persistence\Generic\Mapper\DataMapper')
            ->getDataMap($this->objectType)
            ->getTableName();
    }

    /**
     * @return array
     */
    public function getStoragePageIds()
    {
        return $this->storagePageIds;
    }

    /**
     * @param array $storagePageIds
     * @return $this
     */
    public function setStoragePageIds($storagePageIds)
    {
        $this->storagePageIds = $storagePageIds;

        return $this;
    }

    /**
     * @return int
     */
    public function getLanguageUid()
    {
        return $this->languageUid;
    }

    /**
     * @param int $languageUid
     * @return $this
     */
    public function setLanguageUid($languageUid)
    {
        $this->languageUid = $languageUid;

        return $this;
    }

    /**
     * @return string
     */
    public function getLanguageMode()
    {
        return $this->languageMode;
    }

    /**
     * @param string $languageMode
     * @return $this
     */
    public function setLanguageMode($languageMode)
    {
        $this->languageMode = $languageMode;

        return $this;
    }

    /**
     * @return mixed
     */
    public function getLanguageOverlayMode()
    {
        return $this->languageOverlayMode;
    }

    /**
     * @param mixed $languageOverlayMode
     * @return $this
     */
    public function setLanguageOverlayMode($languageOverlayMode)
    {
        $this->languageOverlayMode = $languageOverlayMode;

        return $this;
    }

    /**
     * @return boolean
     */
    public function getIncludeDeleted()
    {
        return $this->includeDeleted;
    }

    /**
     * @param boolean $includeDeleted
     * @return $this
     */
    public function setIncludeDeleted($includeDeleted)
    {
        $this->includeDeleted = $includeDeleted;

        return $this;
    }

    /**
     * @return boolean
     */
    public function getIgnoreEnableFields()
    {
        return $this->ignoreEnableFields;
    }

    /**
     * @param boolean $ignoreEnableFields
     * @return $this
     */
    public function setIgnoreEnableFields($ignoreEnableFields)
    {
        $this->ignoreEnableFields = $ignoreEnableFields;

        return $this;
    }

    /**
     * @return boolean
     */
    public function getRespectSysLanguage()
    {
        return $this->respectSysLanguage;
    }

    /**
     * @param boolean $respectSysLanguage
     * @return $this
     */
    public function setRespectSysLanguage($respectSysLanguage)
    {
        $this->respectSysLanguage = $respectSysLanguage;

        return $this;
    }

    /**
     * @return boolean
     */
    public function getRespectStoragePage()
    {
        return $this->respectStoragePage;
    }

    /**
     * @param boolean $respectStoragePage
     * @return $this
     */
    public function setRespectStoragePage($respectStoragePage)
    {
        $this->respectStoragePage = $respectStoragePage;

        return $this;
    }

    /**
     * @return array
     */
    public function getEnableFieldsToBeIgnored()
    {
        return $this->enableFieldsToBeIgnored;
    }

    /**
     * @param array $enableFieldsToBeIgnored
     * @return $this
     */
    public function setEnableFieldsToBeIgnored($enableFieldsToBeIgnored)
    {
        $this->enableFieldsToBeIgnored = $enableFieldsToBeIgnored;

        return $this;
    }
}