<?php
/***************************************************************
*  Copyright notice
*
*  (c) 2013 IMIA net based solutions (d.frerich@imia.de)
*  All rights reserved
*
*  This script is part of the TYPO3 project. The TYPO3 project is
*  free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  The GNU General Public License can be found at
*  http://www.gnu.org/copyleft/gpl.html.
*
*  This script is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  This copyright notice MUST APPEAR in all copies of the script!
***************************************************************/

namespace IMIA\ImiaBaseExt\Metadata;

use IMIA\ImiaBaseExt\Annotation\SQL;
use IMIA\ImiaBaseExt\Annotation\TCA;
use IMIA\ImiaBaseExt\Utility\DynamicConfiguration;
use Metadata\MergeableInterface;

/**
 * @package     imia_base_ext
 * @subpackage  Metadata
 * @author      David Frerich <d.frerich@imia.de>
 */
class ClassMetadata extends \Metadata\MergeableClassMetadata
{
    /**
     * @var string
     */
    public $mergeType;

    /**
     * @var Table
     */
    public $sqlTable;

    /**
     * @var Index
     */
    public $sqlIndex;

    /**
     * @var \IMIA\ImiaBaseExt\Annotation\TCA\Table
     */
    public $tcaTable;

    /**
     * @var \IMIA\ImiaBaseExt\Annotation\TCA\AddPaletteToTypes
     */
    public $tcaAddPaletteToTypes;

    /**
     * @var \IMIA\ImiaBaseExt\Annotation\TCA\AddPalettesToTypes
     */
    public $tcaAddPalettesToTypes;

    /**
     * @param MergeableInterface $object
     */
    public function merge(MergeableInterface $object)
    {
        $tcaProperties = array('tcaTable', 'tcaAddPaletteToTypes', 'tcaAddPalettesToTypes');
        $sqlProperties = array('sqlTable', 'sqlIndex');

        $currentExtKey = DynamicConfiguration::getExtKeyForClass($this->name);
        $nextExtKey = DynamicConfiguration::getExtKeyForClass($object->name);

        if ($currentExtKey && $nextExtKey && $currentExtKey !== $nextExtKey) {
            // conflicts detected
            switch ($this->mergeType) {
                case 'sql':
                    // merge sql (table name), don't merge tca
                    foreach ($sqlProperties as $property) {
                        $this->mergeObjectProperty($object, $property);
                    }
                    foreach ($tcaProperties as $property) {
                        if (is_object($object->$property)) {
                            $this->$property = clone $object->$property;
                        } else {
                            $this->$property = $object->$property;
                        }
                    }

                    // replace property metadata
                    foreach ($this->propertyMetadata as $field => $propertyMetaData) {
                        if ($propertyMetaData) {
                            $this->propertyMetadata[$field] = $object->propertyMetadata[$field]
                                ? clone $object->propertyMetadata[$field]
                                : null;
                        }
                    }
                    break;
                case 'tca':
                    // merge sql (table name), don't merge tca
                    foreach ($sqlProperties as $property) {
                        $this->mergeObjectProperty($object, $property);
                    }
                    foreach ($tcaProperties as $property) {
                        if (is_object($object->$property)) {
                            $this->$property = clone $object->$property;
                        } else {
                            $this->$property = $object->$property;
                        }
                    }

                    // replace property metadata
                    foreach ($this->propertyMetadata as $field => $propertyMetaData) {
                        if ($propertyMetaData && is_object($propertyMetaData)) {
                            $sqlColumn = $propertyMetaData->sqlColumn
                                ? clone $propertyMetaData->sqlColumn
                                : null;

                            $this->propertyMetadata[$field] = $object->propertyMetadata[$field]
                                ? clone $object->propertyMetadata[$field]
                                : null;

                            if ($this->propertyMetadata[$field] && is_object($this->propertyMetadata[$field])) {
                                $this->propertyMetadata[$field]->sqlColumn = $sqlColumn;
                            }
                        }
                    }
                    break;
                case 'ts':
                    // merge sql and tca
                    foreach ($sqlProperties as $property) {
                        $this->mergeObjectProperty($object, $property);
                    }
                    foreach ($tcaProperties as $property) {
                        $this->mergeObjectProperty($object, $property);
                    }
                    break;
            }
        } else {
            // merge all, no problems
            foreach ($sqlProperties as $property) {
                $this->mergeObjectProperty($object, $property);
            }
            foreach ($tcaProperties as $property) {
                $this->mergeObjectProperty($object, $property);
            }
        }

        parent::merge($object);
    }

    /**
     * @param MergeableInterface $object
     * @param string $property
     */
    private function mergeObjectProperty(MergeableInterface $object, $property)
    {
        if ($this->$property && $object->$property) {
            foreach (get_object_vars($this->$property) as $objectProperty => $value) {
                if (isset($object->$property->$objectProperty)) {
                    if (is_object($object->$property->$objectProperty)) {
                        $this->$property->$objectProperty = clone $object->$property->$objectProperty;
                    } else {
                        $this->$property->$objectProperty = $object->$property->$objectProperty;
                    }
                }
            }
        } elseif ($object->$property) {
            switch ($property) {
                case 'sqlTable':
                    $this->$property = new SQL\Table;
                    $this->mergeObjectProperty($object, $property);
                    break;
                case 'sqlIndex':
                    $this->$property = new SQL\Index;
                    $this->mergeObjectProperty($object, $property);
                    break;
                case 'tcaTable':
                    $this->$property = new TCA\Table;
                    $this->mergeObjectProperty($object, $property);
                    break;
                case 'tcaAddPaletteToTypes':
                    $this->$property = new TCA\AddPaletteToTypes;
                    $this->mergeObjectProperty($object, $property);
                    break;
                case 'tcaAddPalettesToTypes':
                    $this->$property = new TCA\AddPalettesToTypes;
                    $this->mergeObjectProperty($object, $property);
                    break;
                default:
                    if (is_object($object->$property)) {
                        $this->$property = clone $object->$property;
                    } else {
                        $this->$property = $object->$property;
                    }
                    break;
            }
        }
    }
}
