<?php
/***************************************************************
*  Copyright notice
*
*  (c) 2013 IMIA net based solutions (d.frerich@imia.de)
*  All rights reserved
*
*  This script is part of the TYPO3 project. The TYPO3 project is
*  free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  The GNU General Public License can be found at
*  http://www.gnu.org/copyleft/gpl.html.
*
*  This script is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  This copyright notice MUST APPEAR in all copies of the script!
***************************************************************/

namespace IMIA\ImiaBaseExt\Property;

/**
 * @package     imia_base_ext
 * @subpackage  Property\TypeConverter
 * @author      David Frerich <d.frerich@imia.de>
 */
class PropertyMapper extends \TYPO3\CMS\Extbase\Property\PropertyMapper
{
	/**
	 * Internal function which actually does the property mapping.
	 *
	 * @param mixed $source the source data to map. MUST be a simple type, NO object allowed!
	 * @param string $targetType The type of the target; can be either a class name or a simple type.
	 * @param \TYPO3\CMS\Extbase\Property\PropertyMappingConfigurationInterface $configuration Configuration for the property mapping.
	 * @param array $currentPropertyPath The property path currently being mapped; used for knowing the context in case an exception is thrown.
	 * @throws Exception\TypeConverterException
	 * @return mixed an instance of $targetType
	 */
	protected function doMapping($source, $targetType, \TYPO3\CMS\Extbase\Property\PropertyMappingConfigurationInterface $configuration, &$currentPropertyPath)
    {
		if ($source === NULL) {
			$source = '';
		}
		// This is needed to correctly convert old class names to new ones
		// This compatibility layer will be removed with 7.0
		$targetType = \TYPO3\CMS\Core\Core\ClassLoader::getClassNameForAlias($targetType);
		$typeConverter = $this->findTypeConverter($source, $targetType, $configuration);
		if (!is_object($typeConverter) || !$typeConverter instanceof \TYPO3\CMS\Extbase\Property\TypeConverterInterface) {
			throw new \TYPO3\CMS\Extbase\Property\Exception\TypeConverterException('Type converter for "' . $source . '" -> "' . $targetType . '" not found.');
		}
		$convertedChildProperties = array();
		foreach ($typeConverter->getSourceChildPropertiesToBeConverted($source) as $sourcePropertyName => $sourcePropertyValue) {
			$targetPropertyName = $configuration->getTargetPropertyName($sourcePropertyName);
			if (!$configuration->shouldMap($targetPropertyName)) {
				continue;
			}
			$targetPropertyType = $typeConverter->getTypeOfChildProperty($targetType, $targetPropertyName, $configuration);
			$subConfiguration = $configuration->getConfigurationFor($targetPropertyName);
			$currentPropertyPath[] = $targetPropertyName;
			$targetPropertyValue = $this->doMapping($sourcePropertyValue, $targetPropertyType, $subConfiguration, $currentPropertyPath);
			array_pop($currentPropertyPath);
            // null has to be possible!
			//if ($targetPropertyValue !== NULL) {
				$convertedChildProperties[$targetPropertyName] = $targetPropertyValue;
			//}
		}
		$result = $typeConverter->convertFrom($source, $targetType, $convertedChildProperties, $configuration);
		if ($result instanceof \TYPO3\CMS\Extbase\Error\Error) {
			$this->messages->forProperty(implode('.', $currentPropertyPath))->addError($result);
			$result = NULL;
		}
		return $result;
	}
}