<?php
/***************************************************************
*  Copyright notice
*
*  (c) 2013 IMIA net based solutions (d.frerich@imia.de)
*  All rights reserved
*
*  This script is part of the TYPO3 project. The TYPO3 project is
*  free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  The GNU General Public License can be found at
*  http://www.gnu.org/copyleft/gpl.html.
*
*  This script is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  This copyright notice MUST APPEAR in all copies of the script!
***************************************************************/

namespace IMIA\ImiaBaseExt\Utility\DynamicConfiguration;

use Exception;
use IMIA\ImiaBaseExt\Utility\File;
use IMIA\ImiaBaseExt\Utility\DynamicConfiguration;
use TYPO3\CMS\Core\Utility\GeneralUtility;

/**
 * @package     imia_base_ext
 * @subpackage  Utility\DynamicConfiguration
 * @author      David Frerich <d.frerich@imia.de>
 */
class TCA
{
    /**
     * @var array
     */
    static private $tcaColumnNoAutoLabel = array(
        'IMIA\ImiaBaseExt\Annotation\TCA\None',
        'IMIA\ImiaBaseExt\Annotation\TCA\Passthrough',
    );

    /**
     * @var string
     */
    private $extKey;

    /**
     * @var string
     */
    private $extPath;

    /**
     * @var string
     */
    private $extLangFile;

    /**
     * @var array
     */
    private $run = array();

    /**
     * @var array
     */
    private $runBefore = array();

    /**
     * @param string $extKey
     */
    public function __construct($extKey)
    {
        $this->extKey = $extKey;
        $this->extPath = \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath($extKey);
    }

    /**
     * @param IMIA\ImiaBaseExt\Metadata\ClassMetadata $metadata
     * @param IMIA\ImiaBaseExt\Utility\File $cacheFile
     * @return string
     */
    public function build($metadata, $cacheFile)
    {
        $tcaCacheFiles = array('ext_tables' => $cacheFile);
        foreach (array('DB.xlf', 'db.xlf', 'DB.xml', 'db.xml', 'locallang_db.xml') as $langFile) {
            if (file_exists($this->extPath . 'Resources/Private/Language/' . $langFile)) {
                $this->extLangFile = $langFile;
                break;
            }
        }

        foreach ($metadata as $className => $classMetadata) {
            if ($classMetadata->tcaTable && !$classMetadata->tcaTable->ignore && !$classMetadata->reflection->isAbstract()) {
                $tableName = SQL::getTableName($this->extKey, $classMetadata);
                $realClassName = substr($className, strrpos($className, '\\') + 1);
                if (strpos($realClassName, '_') === false) {
                    $dynTCAFile = $realClassName . '.php';
                } else {
                    $dynTCAFile = GeneralUtility::underscoredToUpperCamelCase($realClassName) . '.php';
                }
                $overwriteCtrl = $this->buildOverwriteCtrl($classMetadata->tcaTable, $tableName);

                if ($classMetadata->tcaTable->title == null) {
                    $classMetadata->tcaTable->title = 'LLL:EXT:' . $this->extKey . '/Resources/Private/Language/' . $this->extLangFile . ':' . File::flatten($realClassName);
                    $this->languageValue($classMetadata->tcaTable->title, $tableName);
                }

                if ($classMetadata->tcaTable->iconfile == null) {
                    foreach (array($tableName . '.png', $tableName . '.gif', $realClassName . '.png', $realClassName . '.gif') as $fileName) {
                        if (file_exists($this->extPath . 'Resources/Public/Images/Icons/' . $fileName)) {
                            $classMetadata->tcaTable->iconfile = 'Resources/Public/Images/Icons/' . $fileName;
                            break;
                        }
                    }
                }

                $searchFields = array();
                foreach ($classMetadata->propertyMetadata as $property => $propertyMetadata) {
                    if ($propertyMetadata->tcaSearchField) {
                        $searchFields[] = SQL::getColumnName($propertyMetadata);
                    }
                }
                if (count($searchFields)) {
                    if ($classMetadata->tcaTable->searchFields && is_array($classMetadata->tcaTable->searchFields)) {
                        $classMetadata->tcaTable->searchFields = GeneralUtility::array_merge_recursive_overrule(
                            $classMetadata->tcaTable->searchFields, $searchFields);
                    } else {
                        $classMetadata->tcaTable->searchFields = $searchFields;
                    }
                }

                $tcaCacheFiles['ext_tables']->addContents(LF .
                    '// ******************************************************************' . LF .
                    '// TCA for ' . $className . LF .
                    '// (' . $tableName . ')' . LF .
                    '// ******************************************************************' . LF);

                if (array_key_exists($tableName, $this->runBefore)) {
                    foreach ($this->runBefore[$tableName] as $run) {
                        $tcaCacheFiles['ext_tables']->addContents($run . ';' . LF);
                    }
                }

                if (strpos($tableName, 'tx_' . File::flatten($this->extKey)) === 0) {
                    // is a extension table (save!)
                    $tcaCacheFiles['ext_tables']->addContents(
                        '$GLOBALS[\'TCA\'][\'' . $tableName . '\'] = array (' . LF .
                        '    \'ctrl\' => array (' . LF .
                                $this->buildCtrl($classMetadata->tcaTable) .
                        '        \'dynamicConfigFile\' => ' . DynamicConfiguration::getExtPath($cacheFile->getPath() . $dynTCAFile) . ',' . LF .
                        '    )' . LF .
                        ');'
                    );
                } else {
                    // could be a table from another extension ore the core (add overwrite if/else)
                    $tcaCacheFiles['ext_tables']->addContents(
                        'if (array_key_exists(\'' . $tableName. '\', $GLOBALS[\'TCA\'])) {' . LF .
                        '    if ((float)TYPO3_version < 6.2) {' . LF .
                        '        \TYPO3\CMS\Core\Utility\GeneralUtility::loadTCA(\'' . $tableName . '\');' . LF .
                        '    }' . LF .
                            $overwriteCtrl .
                            $this->buildNewOverwriteColumns($classMetadata, $tableName, $realClassName) .
                        '} else {' . LF .
                        '    $GLOBALS[\'TCA\'][\'' . $tableName . '\'] = array (' . LF .
                        '        \'ctrl\' => array (' . LF .
                                    $this->buildCtrl($classMetadata->tcaTable, 3) .
                        '            \'dynamicConfigFile\' => ' . DynamicConfiguration::getExtPath($cacheFile->getPath() . $dynTCAFile) . ',' . LF .
                        '        )' . LF .
                        '    );' . LF .
                        '}'
                    );
                }

                if ($classMetadata->tcaTable->allowTableOnStandardPages) {
                    $tcaCacheFiles['ext_tables']->addContents(LF .
                        '\TYPO3\CMS\Core\Utility\ExtensionManagementUtility::allowTableOnStandardPages(\'' . $tableName .  '\');');
                }

                $addToAllTCATypes = array();
                if ($classMetadata->tcaAddPaletteToTypes) {
                    if ($classMetadata->tcaAddPaletteToTypes->label === null) {
                        $classMetadata->tcaAddPaletteToTypes->label = 'LLL:EXT:' . $this->extKey . '/Resources/Private/Language/' .
                            $this->extLangFile . ':' . File::flatten($realClassName) . '_' . File::flatten($classMetadata->tcaAddPaletteToTypes->palette);
                    }

                    $addToAllTCATypes[str_pad((int)$classMetadata->tcaAddPaletteToTypes->priority, 5, '0', STR_PAD_LEFT) . '_' . spl_object_hash($classMetadata)] =
                        '\TYPO3\CMS\Core\Utility\ExtensionManagementUtility::addToAllTCAtypes(' .
                            '\'' . $tableName . '\', ' .
                            '\''. ($classMetadata->tcaAddPaletteToTypes->createTab ? '--div--;' . $this->getReplacedString($classMetadata->tcaAddPaletteToTypes->createTab) . ',' : '') .
                                '--palette--;' . $classMetadata->tcaAddPaletteToTypes->label . ';' . $classMetadata->tcaAddPaletteToTypes->palette . '\', ' .
                            '\'' . $classMetadata->tcaAddPaletteToTypes->types . '\'' .
                            ($classMetadata->tcaAddPaletteToTypes->position ? ', \'' . $classMetadata->tcaAddPaletteToTypes->position . '\'' : '') .
                        ');';
                }
                if ($classMetadata->tcaAddPalettesToTypes) {
                    if ($classMetadata->tcaAddPalettesToTypes->palettes) {
                        foreach ($classMetadata->tcaAddPalettesToTypes->palettes as $palette) {
                            $palette = (object)$palette;
                            if ($palette->label === null) {
                                $palette->label = 'LLL:EXT:' . $this->extKey . '/Resources/Private/Language/' .
                                    $this->extLangFile . ':' . File::flatten($realClassName) . '_' . File::flatten($palette->palette);
                            }

                            $addToAllTCATypes[str_pad((int)$palette->priority, 5, '0', STR_PAD_LEFT) . '_' . spl_object_hash($classMetadata)] =
                                '\TYPO3\CMS\Core\Utility\ExtensionManagementUtility::addToAllTCAtypes(' .
                                    '\'' . $tableName . '\', ' .
                                    '\''. ($palette->createTab ? '--div--;' . $this->getReplacedString($palette->createTab) . ',' : '') .
                                        '--palette--;' . $palette->label . ';' . $palette->palette . '\', ' .
                                    '\'' . $palette->types . '\'' .
                                    ($palette->position ? ', \'' . $palette->position . '\'' : '') .
                                ');';
                        }
                    }
                }

                foreach ($classMetadata->propertyMetadata as $property => $propertyMetadata) {
                    if ($propertyMetadata->tcaAddToTypes && ($propertyMetadata->tcaAddToTypes->heredity || $classMetadata->name == $propertyMetadata->class)) {
                        $addToAllTCATypes[str_pad((int)$propertyMetadata->tcaAddToTypes->priority, 5, '0', STR_PAD_LEFT) . '_' . spl_object_hash($propertyMetadata)] =
                            '\TYPO3\CMS\Core\Utility\ExtensionManagementUtility::addToAllTCAtypes(' .
                                '\'' . $tableName . '\', ' .
                                '\''. ($propertyMetadata->tcaAddToTypes->createTab ? '--div--;' . $this->getReplacedString($propertyMetadata->tcaAddToTypes->createTab) . ',' : '') .
                                    SQL::getColumnName($propertyMetadata) . ($propertyMetadata->tcaAddToTypes->options ?: '') . '\', ' .
                                '\'' . $propertyMetadata->tcaAddToTypes->types . '\'' .
                                ($propertyMetadata->tcaAddToTypes->position ? ', \'' . $propertyMetadata->tcaAddToTypes->position . '\'' : '') .
                            ');'
                        ;
                    }
                }

                $addToAllTCATypesStr = '';
                if (count($addToAllTCATypes) > 0) {
                    ksort($addToAllTCATypes);
                    $addToAllTCATypesStr = LF . implode(LF, $addToAllTCATypes);
                    $tcaCacheFiles['ext_tables']->addContents($addToAllTCATypesStr);
                }

                $tcaCacheFiles['ext_tca_' . File::flatten($realClassName)] =
                    new File($dynTCAFile, $cacheFile->getPath(),
                        $this->buildDynFile($classMetadata, $tableName, $realClassName, $addToAllTCATypesStr));

                if (array_key_exists($tableName, $this->run)) {
                    foreach ($this->run[$tableName] as $run) {
                        $tcaCacheFiles['ext_tables']->addContents(LF . $run . ';');
                    }
                }
            }
        }

        return $tcaCacheFiles;
    }

    /**
     * @param \IMIA\ImiaBaseExt\Annotation\TCA\Table $tcaTable
     * @return string
     */
    private function buildCtrl($tcaTable, $indent = 2)
    {
        $ctrl = '';
        foreach (get_object_vars($tcaTable) as $property => $value) {
            if (!in_array($property, array('types', 'palettes', 'showRecordFieldList', 'allowTableOnStandardPages', 'ignore'))) {
                if ($value !== null || array_key_exists($property, $tcaTable::$defaults)) {
                    if ($value === null) {
                        $value = $tcaTable::$defaults[$property];
                    }
                    if (in_array($property, $tcaTable::$implodes)) {
                        $value = implode(',', $value);
                    }

                    $phpValue = $this->getValueForPhp($value, 4);
                    if ($property && $property == 'iconfile') {
                        $phpValue = '\TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extRelPath(' .
                            $this->getValueForPhp($this->extKey) . ') . ' . $phpValue;
                    }

                    $ctrl .= str_pad('', $indent*4) . '\'' . $property . '\' => ' . $phpValue . ',' . LF;
                }
            }
        }

        return $ctrl;
    }

    /**
     * @param \IMIA\ImiaBaseExt\Annotation\TCA\Table $tcaTable
     * @param string $tableName
     * @return string
     */
    private function buildOverwriteCtrl($tcaTable, $tableName)
    {
        $ctrl = '';
        foreach (get_object_vars($tcaTable) as $property => $value) {
            if (!in_array($property, array('showRecordFieldList', 'allowTableOnStandardPages', 'ignore'))) {
                if ($value !== null) {
                    if (in_array($property, $tcaTable::$implodes)) {
                        $value = implode(',', $value);
                    }

                    $propertyVariable = '$GLOBALS[\'TCA\'][\'' . $tableName . '\']';
                    if (!in_array($property, array('types', 'palettes', 'showRecordFieldList'))) {
                        $propertyVariable .= '[\'ctrl\']';
                    }

                    if (in_array($property, array('types', 'palettes'))) {
                        $newValue = array();
                        foreach ($value as $key => $realValues) {
                            if ($property == 'palettes') {
                                $newValue[(string)$key] = array('showitem' => $realValues, 'canNotCollapse' => true);
                            } else{
                                $newValue[(string)$key] = array('showitem' => $realValues);
                            }
                        }
                        $value = $newValue;
                    }

                    $ctrl .= $this->buildOverwriteProperty($propertyVariable, $property, $value);
                }
            }
        }

        return $ctrl;
    }

    /**
     * @param string $propertyVariable
     * @param string $property
     * @param mixed $value
     * @param integer $indent
     * @return string
     */
    private function buildOverwriteProperty($propertyVariable, $property, $value, $indent = 1)
    {
        $propertyCode = '';
        $phpValue = $this->getValueForPhp($value, 4);

        if ($property && $property == 'iconfile') {
            $phpValue = '\TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extRelPath(' .
                $this->getValueForPhp($this->extKey) . ') . ' . $phpValue;
        }

        if (is_array($value)) {
            $setValues = '';
            foreach ($value as $key => $val) {
                $setValues .= $this->buildOverwriteProperty($propertyVariable . '[\'' . $property . '\']', $key, $val, $indent+1);
            }

            $propertyCode .=
                str_pad('', $indent*4) . 'if (array_key_exists(\'' . $property . '\', ' . $propertyVariable . ')) {' . LF .
                $setValues .
                str_pad('', $indent*4) . '} else {' . LF .
                str_pad('', ($indent+1)*4) . $propertyVariable . '[\'' . $property . '\'] = ' . $phpValue . ';' . LF .
                str_pad('', $indent*4) . '}' . LF;
        } else {
            $propertyCode .= str_pad('', $indent*4) . $propertyVariable . '[\'' . $property . '\'] = ' . $phpValue . ';' . LF;
        }

        return $propertyCode;
    }

    /**
     * @param IMIA\ImiaBaseExt\Metadata\ClassMetadata $metadata
     * @param string $tableName
     * @param string $realClassName
     * @return string
     */
    private function buildNewOverwriteColumns($classMetadata, $tableName, $realClassName)
    {
        $columns = '';
        foreach ($classMetadata->propertyMetadata as $property => $propertyMetadata) {
            if ($propertyMetadata->tcaColumn) {
                $name = SQL::getColumnName($propertyMetadata);

                if ($propertyMetadata->tcaColumn->label == '[' . $name . ']') {
                    $propertyMetadata->tcaColumn->label = null;
                }
                $overwriteColumn = $this->buildOverwriteColumn($propertyMetadata, $tableName);

                if ($propertyMetadata->tcaColumn->label === null) {
                    if (!in_array(get_class($propertyMetadata->tcaColumn), self::$tcaColumnNoAutoLabel)) {
                        $propertyMetadata->tcaColumn->label = 'LLL:EXT:' . $this->extKey . '/Resources/Private/Language/' . $this->extLangFile . ':' .
                            File::flatten($realClassName) . '_' . File::flatten($property);
                    }
                    $langTest = $this->languageValue($propertyMetadata->tcaColumn->label, $name);

                    if (!$langTest) {
                        $reflection = $classMetadata->reflection;
                        while ($reflection->getParentClass() && !$langTest) {
                            $reflection = $reflection->getParentClass();
                            $parentRealClassName = substr($reflection->name, strrpos($reflection->name, '\\') + 1);

                            $propertyMetadata->tcaColumn->label = 'LLL:EXT:' . $this->extKey . '/Resources/Private/Language/' . $this->extLangFile . ':' .
                                File::flatten($parentRealClassName) . '_' . File::flatten($property);
                            $langTest = $this->languageValue($propertyMetadata->tcaColumn->label, $name);
                        }
                    }
                }

                $columns .=
                    '    if (array_key_exists(\'' . $name . '\', $GLOBALS[\'TCA\'][\'' . $tableName . '\'][\'columns\'])) {' . LF .
                            $overwriteColumn .
                    '    } else {' . LF .
                    '        $tempColumn = array(' . LF .
                                $this->buildColumn($propertyMetadata, $tableName, 3) .
                    '        );' . LF .
                    '        \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::addTCAcolumns(\'' . $tableName .  '\', $tempColumn, 1);' . LF .
                    '    }' . LF;
            }
        }

        return $columns;
    }

    /**
     * @param IMIA\ImiaBaseExt\Metadata\ClassMetadata $metadata
     * @param string $tableName
     * @param string $realClassName
     * @param string $addToAllTCATypesStr
     * @return string
     */
    private function buildDynFile($classMetadata, $tableName, $realClassName, $addToAllTCATypesStr = '')
    {
        $tca = '<?php' . LF .
            'if (!defined (\'TYPO3_MODE\')) {' . LF .
            '	die (\'Access denied.\');' . LF .
            '}' . LF .
            LF .
            '$TCA[\'' . $tableName . '\'] = array(' . LF .
            '    \'ctrl\' => $TCA[\'' . $tableName . '\'][\'ctrl\'],' . LF .
            '    \'interface\' => array(' . LF .
            '        \'showRecordFieldList\' => ' .  $this->buildShowRecordFieldList($classMetadata) . ',' . LF .
            '    ),' . LF .
            '    \'types\' => ' . $this->buildTypes($classMetadata) . ',' . LF .
            '    \'palettes\' => ' . $this->buildPalettes($classMetadata) . ',' . LF .
            '    \'columns\' => array(' . LF;

        foreach ($classMetadata->propertyMetadata as $property => $propertyMetadata) {
            if ($propertyMetadata->tcaColumn) {
                $name = SQL::getColumnName($propertyMetadata);
                if ($propertyMetadata->tcaColumn->label == '[' . $name . ']') {
                    $propertyMetadata->tcaColumn->label = null;
                }

                if ($propertyMetadata->tcaColumn->label === null) {
                    if (!in_array(get_class($propertyMetadata->tcaColumn), self::$tcaColumnNoAutoLabel)) {
                        $propertyMetadata->tcaColumn->label = 'LLL:EXT:' . $this->extKey . '/Resources/Private/Language/' . $this->extLangFile . ':' .
                            File::flatten($realClassName) . '_' . File::flatten($property);
                    }
                    $this->languageValue($propertyMetadata->tcaColumn->label, $name);
                }

                $tca .= $this->buildColumn($propertyMetadata, $tableName);
            }
        }

        $tca .=
            '   ),' . LF .
            ');' . $addToAllTCATypesStr;

        return $tca;
    }

    /**
     * @param IMIA\ImiaBaseExt\Metadata\ClassMetadata $classMetadata
     * @return string
     */
    private function buildShowRecordFieldList($classMetadata)
    {
        $tcaTable = $classMetadata->tcaTable;
        if ($tcaTable->showRecordFieldList !== null) {
            $showRecordFieldList = $tcaTable->showRecordFieldList;
        } else {
            $columns = $this->getColumnsToShow($classMetadata);
            $showRecordFieldList = implode(', ', $columns);
        }

        return $this->getValueForPhp($showRecordFieldList);
    }

    /**
     * @param IMIA\ImiaBaseExt\Metadata\ClassMetadata $classMetadata
     * @return string
     */
    private function buildTypes($classMetadata)
    {
        $types = array();
        if ($classMetadata->tcaTable->types) {
            foreach ($classMetadata->tcaTable->types as $key => $type) {
                $types[(string)$key] = array('showitem' => $type);
            }
        } else {
            $types[] = array('showitem' => implode(', ', $this->getColumnsToShow($classMetadata)));
        }

        return $this->getValueForPhp($types, 2);
    }

    /**
     * @param IMIA\ImiaBaseExt\Metadata\ClassMetadata $classMetadata
     * @return string
     */
    private function buildPalettes($classMetadata)
    {
        $palettes = array();
        if ($classMetadata->tcaTable->palettes) {
            foreach ($classMetadata->tcaTable->palettes as $key => $palette) {
                $palettes[(string)$key] = array('showitem' => $palette, 'canNotCollapse' => true);
            }
        }

        return $this->getValueForPhp($palettes, 2);
    }

    /**
     * @param IMIA\ImiaBaseExt\Metadata\ClassMetadata $classMetadata
     * @return array
     */
    private function getColumnsToShow($classMetadata)
    {
        $columns = array();
        foreach ($classMetadata->propertyMetadata as $property => $propertyMetadata) {
            $column = $propertyMetadata->tcaColumn;
            if ($column && get_class($column) !== 'IMIA\ImiaBaseExt\Annotation\TCA\Passthrough') {
                $columns[] = SQL::getColumnName($propertyMetadata);
            }
        }

        return $columns;
    }

    /**
     * @param IMIA\ImiaBaseExt\Metadata\PropertyMetadata $propertyMetadata
     * @param string $tableName
     * @return string
     */
    private function buildColumn($propertyMetadata, $tableName, $indent = 2)
    {
        $column = $propertyMetadata->tcaColumn;
        $name = SQL::getColumnName($propertyMetadata);

        $tca = str_pad('', $indent * 4) . '\'' . $name . '\' => array(' . LF;

        $config = array();
        foreach (get_object_vars($column) as $property => $value) {
            if (array_key_exists($property, $column::$fixed)) {
                $value = $column::$fixed[$property];
            }

            if ($value !== null || array_key_exists($property, $column::$defaults)) {
                if ($value === null) {
                    $value = $column::$defaults[$property];
                } elseif (is_array($value) && array_key_exists($property, $column::$defaults) && is_array($column::$defaults[$property])) {
                    $value = GeneralUtility::array_merge_recursive_overrule($column::$defaults[$property], $value);
                }

                if (in_array($property, $column::$implodes)) {
                    $value = implode(',', $value);
                }

                if (is_string($property)) {
                    $value = $this->replaceAdditionalMarkers($value, $name, $tableName);
                }

                if ($property == 'items' && !is_array($value)) {
                    $items = null;
                    try {
                        @eval('$items = ' . $value . ';');
                    } catch(Exception $e){ }

                    $value = array();
                    if ($items && is_array($items)) {
                        foreach ($items as $key => $item) {
                            $icon = 'none';
                            if (is_array($item)) {
                                $icon = $item[1];
                                $item = $item[0];
                            }
                            $value[] = array((string)$item, $key, $icon);
                        }
                    }
                }

                if (in_array($property, $column::$config)) {
                    $config[$property] = $value;
                } else {
                    $tca .= str_pad('', ($indent+1) * 4) . '\'' . $property . '\' => ' . $this->getValueForPhp($value, $indent+2) . ',' . LF;
                }
            }
        }

        if ($config['items'] && is_array($config['items'])
          && $config['renderMode'] != 'checkbox' && (int)$config['maxItems'] <= 1) {
            foreach ($config['items'] as $key => $val) {
                if (count($val) > 2) {
                    array_pop($config['items'][$key]);
                }
            }
        }

        $configValue = $this->getValueForPhp($config, $indent+2);
        if (property_exists($column, 'configFunction')) {
            $param = '';
            if (property_exists($column, 'configParameter')) {
                $param = ', \'' . $column::$configParameter . '\'';
            }

            $configValue = '\TYPO3\CMS\Core\Utility\ExtensionManagementUtility::' . $column::$configFunction . '(\'' . $name . '\', ' . $configValue . $param . ')';
        }

        if (property_exists($column, 'runBefore')) {
            $this->runBefore[$tableName][$name] = $this->replaceAdditionalMarkers($column::$runBefore, $name, $tableName);
        }
        if (property_exists($column, 'run')) {
            $this->run[$tableName][$name] = $this->replaceAdditionalMarkers($column::$run, $name, $tableName);
        }

        $tca .= str_pad('', ($indent+1) * 4) . '\'config\' => ' . $configValue . ',' . LF;
        $tca .= str_pad('', $indent * 4) . '),' . LF;

        return $tca;
    }

    /**
     * @param IMIA\ImiaBaseExt\Metadata\PropertyMetadata $propertyMetadata
     * @param string $tableName
     * @return string
     */
    private function buildOverwriteColumn($propertyMetadata, $tableName)
    {
        $column = $propertyMetadata->tcaColumn;
        $name = SQL::getColumnName($propertyMetadata);

        $columnVariable = '$GLOBALS[\'TCA\'][\'' . $tableName . '\'][\'columns\'][\'' . $name . '\']';

        $config = array();
        $tca = '';
        foreach (get_object_vars($column) as $property => $value) {
            if (array_key_exists($property, $column::$fixed)) {
                $value = $column::$fixed[$property];
            }

            if ($property == 'defaultExtras' && $value === null) {
                $value = $column::$defaults[$property];
            }

            if ($value !== null) {
                if (in_array($property, $column::$implodes)) {
                    $value = implode(',', $value);
                }
                if (is_string($property)) {
                    $value = str_replace(array('###CURRENT_TABLE###'), array($tableName), $value);
                }

                if (in_array($property, $column::$config)) {
                    $config[$property] = $value;
                } else {
                    $tca .= $this->buildOverwriteProperty($columnVariable, $property, $value, 2);
                }
            }
        }

        if (property_exists($column, 'configFunction')) {
            $tca .= '        ' . $columnVariable . '[\'config\'] = \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::' .
                $column::$configFunction . '(\'' . $name . '\', ' . $this->getValueForPhp($config, 3) . ');' . LF;
        } else {
            $tca .= $this->buildOverwriteProperty($columnVariable, 'config', $config, 2);
        }

        return $tca;
    }

    /**
     * @param mixed $value
     * @param integer $indent
     * @return string
     */
    private function getValueForPhp($value, $indent = 0)
    {
        if (is_string($value)) {
            // enable php code (e.g. for dates range: mktime)
            if (preg_match('/^<\?(php)? (.+)[;]? \?>$/ism', trim($value), $match)) {
                $newValue = $match[2];
            } else {
                $value = $this->getReplacedString($value);
                $newValue = '\'' . str_replace(array("'", '\\'), array("\'", '\\'), $value) . '\'';
            }
        } elseif (is_bool($value)) {
            $newValue = $value ? 'true' : 'false';
        } elseif (is_array($value)) {
            $newValue = 'array(' . LF;
            foreach ($value as $key => $keyVal) {
                $newValue .= str_pad('', $indent*4) . $this->getValueForPhp($key) . ' => ' .
                    $this->getValueForPhp($keyVal, $indent+1) . ',' . LF;
            }
            $newValue .= str_pad('', ($indent-1)*4) . ')';
        } else {
            $newValue = $value;
        }

        return $newValue;
    }

    /**
     * @param string $string
     * @return string
     */
    private function getReplacedString($string)
    {
        // replace "LLL::" with default ' . $this->extLangFile . ' path
        if (preg_match_all('/LLL::([^,]+)/ism', trim($string), $matches)) {
            foreach ($matches[1] as $key => $match) {
                $string = str_replace($matches[0][$key], 'LLL:EXT:' . $this->extKey . '/Resources/Private/Language/' . $this->extLangFile . ':' . $match, $string);
            }
        // replace "EXTREL::" with extRelPath
        } elseif (preg_match_all('/EXTREL::([^,]+)/ism', trim($string), $matches)) {
            foreach ($matches[1] as $key => $match) {
                $string = str_replace($matches[0][$key], \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extRelPath($this->extKey) . $match, $string);
            }
        // replace "EXT::" with extPath
        } elseif (preg_match_all('/EXT::([^,]+)/ism', trim($string), $matches)) {
            foreach ($matches[1] as $key => $match) {
                $string = str_replace($matches[0][$key], \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath($this->extKey) . $match, $string);
            }
        }

        return $string;
    }

    /**
     * @param string $value
     * @param string $name
     * @param string $tableName
     * @return string
     */
    private function replaceAdditionalMarkers($value, $name, $tableName)
    {
        return str_replace(array('###CURRENT_TABLE###', '###CURRENT_COLUMN###', '###EXTKEY###'), array($tableName, $name, $this->extKey), $value);
    }

    /**
     * @param string $langString
     * @return string
     */
    private function languageValue(&$langString, $fieldName)
    {
        if ($GLOBALS['LANG']) {
            if ($langString === null || $GLOBALS['LANG']->sL($langString) === '') {
                $langString = '[' . $fieldName . ']';

                return false;
            }

            if ($GLOBALS['LANG']->sL($langString) === 'none') {
                $langString = '';
            }
        }

        return true;
    }
}