<?php
/***************************************************************
*  Copyright notice
*
*  (c) 2013 IMIA net based solutions (d.frerich@imia.de)
*  All rights reserved
*
*  This script is part of the TYPO3 project. The TYPO3 project is
*  free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  The GNU General Public License can be found at
*  http://www.gnu.org/copyleft/gpl.html.
*
*  This script is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  This copyright notice MUST APPEAR in all copies of the script!
***************************************************************/

namespace IMIA\ImiaBaseExt\Xclass\Backend\Form;

use TYPO3\CMS\Backend\Form\Element\InlineElement;
use TYPO3\CMS\Backend\Utility\BackendUtility;
use TYPO3\CMS\Backend\Utility\IconUtility;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Core\Utility\MathUtility;

/**
 * @package     imia_base_ext
 * @subpackage  ViewHelpers
 * @author      David Frerich <d.frerich@imia.de>
 */
class FormEngine extends \TYPO3\CMS\Backend\Form\FormEngine
{
	/**
	 * @param string $icon
	 * @param string $alt
	 * @param string $title
	 * @return string
	 */
	protected function getIconHtml($icon, $alt = '', $title = '')
    {
        if ($icon == 'none') {
            return '';
        } else {
            $iconArray = $this->getIcon($icon);
            if (is_file(GeneralUtility::resolveBackPath(PATH_typo3 . PATH_typo3_mod . $iconArray[0]))) {
                return '<img src="' . $iconArray[0] . '" alt="' . $alt . '" ' . ($title ? 'title="' . $title . '"' : '') . ' />';
            } else {
                return IconUtility::getSpriteIcon($icon, array('alt' => $alt, 'title' => $title));
            }
        }
	}

    /**
     * @inheritdoc
     */
    public function wrapTotal($c, $rec, $table)
    {
        $parts = $this->replaceTableWrap(explode('|', $this->totalWrap, 2), $rec, $table);

        // Hook: getWrapTotal_preProcess
        foreach ($this->hookObjectsMainFields as $hookObj) {
            if (method_exists($hookObj, 'getWrapTotal_preProcess')) {
                $hookObj->getWrapTotal_preProcess($c, $rec, $table, $parts, $this);
            }
        }

        return $parts[0] . $c . $parts[1] . implode('', $this->hiddenFieldAccum);
    }

    /**
     * Returns the form HTML code for a database table field.
     *
     * @param string $table The table name
     * @param string $field The field name
     * @param array $row The record to edit from the database table.
     * @param string $altName Alternative field name label to show.
     * @param boolean $palette Set this if the field is on a palette (in top frame), otherwise not. (if set, field will render as a hidden field).
     * @param string $extra The "extra" options from "Part 4" of the field configurations found in the "types" "showitem" list. Typically parsed by $this->getSpecConfFromString() in order to get the options as an associative array.
     * @param integer $pal The palette pointer.
     * @return mixed String (normal) or array (palettes)
     * @todo Define visibility
     */
    public function getSingleField($table, $field, $row, $altName = '', $palette = 0, $extra = '', $pal = 0) {
        // Hook: getSingleField_preProcess
        foreach ($this->hookObjectsSingleField as $hookObj) {
            if (method_exists($hookObj, 'getSingleField_preProcess')) {
                $hookObj->getSingleField_preProcess($table, $field, $row, $altName, $palette, $extra, $pal, $this);
            }
        }
        $out = '';
        $PA = array();
        $PA['altName'] = $altName;
        $PA['palette'] = $palette;
        $PA['extra'] = $extra;
        $PA['pal'] = $pal;
        // Get the TCA configuration for the current field:
        $PA['fieldConf'] = $GLOBALS['TCA'][$table]['columns'][$field];
        $PA['fieldConf']['config']['form_type'] = $PA['fieldConf']['config']['form_type'] ?: $PA['fieldConf']['config']['type'];

        // Using "form_type" locally in this script
        $skipThisField = $this->inline->skipField($table, $field, $row, $PA['fieldConf']['config']);

        // Evaluate display condition
        $displayConditionResult = TRUE;
        if (is_array($PA['fieldConf']) && $PA['fieldConf']['displayCond'] && is_array($row)) {
            /** @var $elementConditionMatcher \TYPO3\CMS\Backend\Form\ElementConditionMatcher */
            $elementConditionMatcher = GeneralUtility::makeInstance('TYPO3\\CMS\\Backend\\Form\\ElementConditionMatcher');

            /** ### ADDITION by this Xclass ### */
            $elementConditionMatcher->setRecordTable($table);
            /** ### END ADDITION ### */

            $displayConditionResult = $elementConditionMatcher->match($PA['fieldConf']['displayCond'], $row);
        }
        // Check if this field is configured and editable (according to excludefields + other configuration)
        if (
            is_array($PA['fieldConf'])
            && !$skipThisField
            && (!$PA['fieldConf']['exclude'] || $GLOBALS['BE_USER']->check('non_exclude_fields', $table . ':' . $field))
            && $PA['fieldConf']['config']['form_type'] != 'passthrough'
            && ($this->RTEenabled || !$PA['fieldConf']['config']['showIfRTE'])
            && $displayConditionResult
            && (!$GLOBALS['TCA'][$table]['ctrl']['languageField'] || $PA['fieldConf']['l10n_display'] || ($PA['fieldConf']['l10n_mode'] !== 'exclude') || $row[$GLOBALS['TCA'][$table]['ctrl']['languageField']] <= 0)
            && (!$GLOBALS['TCA'][$table]['ctrl']['languageField'] || !$this->localizationMode || $this->localizationMode === $PA['fieldConf']['l10n_cat'])
        ) {
            // Fetching the TSconfig for the current table/field. This includes the $row which means that
            $PA['fieldTSConfig'] = $this->setTSconfig($table, $row, $field);
            // If the field is NOT disabled from TSconfig (which it could have been) then render it
            if (!$PA['fieldTSConfig']['disabled']) {
                // Override fieldConf by fieldTSconfig:
                $PA['fieldConf']['config'] = $this->overrideFieldConf($PA['fieldConf']['config'], $PA['fieldTSConfig']);
                // Init variables:
                $PA['itemFormElName'] = $this->prependFormFieldNames . '[' . $table . '][' . $row['uid'] . '][' . $field . ']';
                // Form field name, in case of file uploads
                $PA['itemFormElName_file'] = $this->prependFormFieldNames_file . '[' . $table . '][' . $row['uid'] . '][' . $field . ']';
                // Form field name, to activate elements
                // If the "eval" list contains "null", elements can be deactivated which results in storing NULL to database
                $PA['itemFormElNameActive'] = $this->prependFormFieldNamesActive . '[' . $table . '][' . $row['uid'] . '][' . $field . ']';
                // The value to show in the form field.
                $PA['itemFormElValue'] = $row[$field];
                $PA['itemFormElID'] = $this->prependFormFieldNames . '_' . $table . '_' . $row['uid'] . '_' . $field;
                // Set field to read-only if configured for translated records to show default language content as readonly
                if ($PA['fieldConf']['l10n_display'] && GeneralUtility::inList($PA['fieldConf']['l10n_display'], 'defaultAsReadonly') && $row[$GLOBALS['TCA'][$table]['ctrl']['languageField']] > 0) {
                    $PA['fieldConf']['config']['readOnly'] = TRUE;
                    $PA['itemFormElValue'] = $this->defaultLanguageData[$table . ':' . $row['uid']][$field];
                }
                if (strpos($GLOBALS['TCA'][$table]['ctrl']['type'], ':') === FALSE) {
                    $typeField = $GLOBALS['TCA'][$table]['ctrl']['type'];
                } else {
                    $typeField = substr($GLOBALS['TCA'][$table]['ctrl']['type'], 0, strpos($GLOBALS['TCA'][$table]['ctrl']['type'], ':'));
                }
                // Create a JavaScript code line which will ask the user to save/update the form due to changing the element. This is used for eg. "type" fields and others configured with "requestUpdate"
                if (
                    !empty($GLOBALS['TCA'][$table]['ctrl']['type'])
                    && $field === $typeField
                    || !empty($GLOBALS['TCA'][$table]['ctrl']['requestUpdate'])
                    && GeneralUtility::inList(str_replace(' ', '', $GLOBALS['TCA'][$table]['ctrl']['requestUpdate']), $field)
                ) {
                    if ($GLOBALS['BE_USER']->jsConfirmation(1)) {
                        $alertMsgOnChange = 'if (confirm(TBE_EDITOR.labels.onChangeAlert) && TBE_EDITOR.checkSubmit(-1)){ TBE_EDITOR.submitForm() };';
                    } else {
                        $alertMsgOnChange = 'if (TBE_EDITOR.checkSubmit(-1)){ TBE_EDITOR.submitForm() };';
                    }
                } else {
                    $alertMsgOnChange = '';
                }
                // Render as a hidden field?
                if (in_array($field, $this->hiddenFieldListArr)) {
                    $this->hiddenFieldAccum[] = '<input type="hidden" name="' . $PA['itemFormElName'] . '" value="' . htmlspecialchars($PA['itemFormElValue']) . '" />';
                } else {
                    // Render as a normal field:
                    // If the field is NOT a palette field, then we might create an icon which links to a palette for the field, if one exists.
                    if (!$PA['palette']) {
                        $paletteFields = $this->loadPaletteElements($table, $row, $PA['pal']);
                        if ($PA['pal'] && $this->isPalettesCollapsed($table, $PA['pal']) && count($paletteFields)) {
                            list($thePalIcon, $palJSfunc) = $this->wrapOpenPalette(IconUtility::getSpriteIcon('actions-system-options-view', array('title' => htmlspecialchars($this->getLL('l_moreOptions')))), $table, $row, $PA['pal'], 1);
                        } else {
                            $thePalIcon = '';
                            $palJSfunc = '';
                        }
                    }
                    // onFocus attribute to add to the field:
                    $PA['onFocus'] = $palJSfunc && !$GLOBALS['BE_USER']->uc['dontShowPalettesOnFocusInAB'] ? ' onfocus="' . htmlspecialchars($palJSfunc) . '"' : '';
                    // Find item
                    $item = '';
                    $PA['label'] = $PA['altName'] ?: $PA['fieldConf']['label'];
                    $PA['label'] = $PA['fieldTSConfig']['label'] ?: $PA['label'];
                    $PA['label'] = $PA['fieldTSConfig']['label.'][$GLOBALS['LANG']->lang] ?: $PA['label'];
                    $PA['label'] = $this->sL($PA['label']);
                    // JavaScript code for event handlers:
                    $PA['fieldChangeFunc'] = array();
                    $PA['fieldChangeFunc']['TBE_EDITOR_fieldChanged'] = 'TBE_EDITOR.fieldChanged(\'' . $table . '\',\'' . $row['uid'] . '\',\'' . $field . '\',\'' . $PA['itemFormElName'] . '\');';
                    $PA['fieldChangeFunc']['alert'] = $alertMsgOnChange;
                    // If this is the child of an inline type and it is the field creating the label
                    if ($this->inline->isInlineChildAndLabelField($table, $field)) {
                        $inlineObjectId = implode(InlineElement::Structure_Separator, array(
                            $this->inline->inlineNames['object'],
                            $table,
                            $row['uid']
                        ));
                        $PA['fieldChangeFunc']['inline'] = 'inline.handleChangedField(\'' . $PA['itemFormElName'] . '\',\'' . $inlineObjectId . '\');';
                    }
                    // Based on the type of the item, call a render function:
                    $item = $this->getSingleField_SW($table, $field, $row, $PA);
                    // Add language + diff
                    if ($PA['fieldConf']['l10n_display'] && (GeneralUtility::inList($PA['fieldConf']['l10n_display'], 'hideDiff') || GeneralUtility::inList($PA['fieldConf']['l10n_display'], 'defaultAsReadonly'))) {
                        $renderLanguageDiff = FALSE;
                    } else {
                        $renderLanguageDiff = TRUE;
                    }
                    if ($renderLanguageDiff) {
                        $item = $this->renderDefaultLanguageContent($table, $field, $row, $item);
                        $item = $this->renderDefaultLanguageDiff($table, $field, $row, $item);
                    }
                    // If the record has been saved and the "linkTitleToSelf" is set, we make the field name into a link, which will load ONLY this field in alt_doc.php
                    $label = htmlspecialchars($PA['label'], ENT_COMPAT, 'UTF-8', FALSE);
                    if (MathUtility::canBeInterpretedAsInteger($row['uid']) && $PA['fieldTSConfig']['linkTitleToSelf'] && !GeneralUtility::_GP('columnsOnly')) {
                        $lTTS_url = $this->backPath . 'alt_doc.php?edit[' . $table . '][' . $row['uid'] . ']=edit&columnsOnly=' . $field . '&returnUrl=' . rawurlencode($this->thisReturnUrl());
                        $label = '<a href="' . htmlspecialchars($lTTS_url) . '">' . $label . '</a>';
                    }

                    if (isset($PA['fieldConf']['config']['mode']) && $PA['fieldConf']['config']['mode'] == 'useOrOverridePlaceholder') {
                        $placeholder = $this->getPlaceholderValue($table, $field, $PA['fieldConf']['config'], $row);
                        $onChange = 'typo3form.fieldTogglePlaceholder(' . GeneralUtility::quoteJSvalue($PA['itemFormElName']) . ', !this.checked)';
                        $checked = $PA['itemFormElValue'] === NULL ? '' : ' checked="checked"';

                        $this->additionalJS_post[] = 'typo3form.fieldTogglePlaceholder('
                            . GeneralUtility::quoteJSvalue($PA['itemFormElName']) . ', ' . ($checked ? 'false' : 'true') . ');';

                        $item = '<div class="t3-form-field-placeholder-override">'
                            . '<span class="t3-tceforms-placeholder-override-checkbox">' .
                            '<input type="hidden" name="' . htmlspecialchars($PA['itemFormElNameActive']) . '" value="0" />' .
                            '<input type="checkbox" name="' . htmlspecialchars($PA['itemFormElNameActive']) . '" value="1" id="tce-forms-textfield-use-override-' . $field . '-' . $row['uid'] . '" onchange="' . htmlspecialchars($onChange) . '"' . $checked . ' />' .
                            '<label for="tce-forms-textfield-use-override-' . $field . '-' . $row['uid'] . '">' .
                            sprintf($GLOBALS['LANG']->sL('LLL:EXT:lang/locallang_core.xlf:labels.placeholder.override'),
                                BackendUtility::getRecordTitlePrep($placeholder, 20)) . '</label>' .
                            '</span>'
                            . '<div class="t3-form-placeholder-placeholder">' . $this->getSingleField_typeNone_render(
                                $PA['fieldConf']['config'], GeneralUtility::fixed_lgd_cs($placeholder, 30)
                            ) . '</div>'
                            . '<div class="t3-form-placeholder-formfield">' . $item . '</div>'
                            . '</div>';
                    }

                    // Wrap the label with help text
                    $PA['label'] = ($label = BackendUtility::wrapInHelp($table, $field, $label));
                    // Create output value:

                    if ((float)TYPO3_version >= 6.2) {
                        $isNull = $this->isDisabledNullValueField($table, $field, $row, $PA);
                    } else {
                        $isNull = $this->isNullValue($table, $field, $row, $PA);
                    }
                    if ($PA['fieldConf']['config']['form_type'] == 'user' && $PA['fieldConf']['config']['noTableWrapping']) {
                        $out = $item;
                    } elseif ($PA['palette']) {
                        // Array:
                        $out = array(
                            'NAME' => $label,
                            'ID' => $row['uid'],
                            'FIELD' => $field,
                            'TABLE' => $table,
                            'ITEM' => $item,
                            'ITEM_DISABLED' => ($isNull ? ' disabled' : ''),
                            'ITEM_NULLVALUE' => $this->renderNullValueWidget($table, $field, $row, $PA),
                        );
                        $out = $this->addUserTemplateMarkers($out, $table, $field, $row, $PA);
                    } else {
                        // String:
                        $out = array(
                            'NAME' => $label,
                            'ITEM' => $item,
                            'TABLE' => $table,
                            'ID' => $row['uid'],
                            'PAL_LINK_ICON' => $thePalIcon,
                            'FIELD' => $field,
                            'ITEM_DISABLED' => ($isNull ? ' disabled' : ''),
                            'ITEM_NULLVALUE' => $this->renderNullValueWidget($table, $field, $row, $PA),
                        );
                        $out = $this->addUserTemplateMarkers($out, $table, $field, $row, $PA);
                        // String:
                        $out = $this->intoTemplate($out);
                    }
                }
            } else {
                $this->commentMessages[] = $this->prependFormFieldNames . '[' . $table . '][' . $row['uid'] . '][' . $field . ']: Disabled by TSconfig';
            }
        }
        // Hook: getSingleField_postProcess
        foreach ($this->hookObjectsSingleField as $hookObj) {
            if (method_exists($hookObj, 'getSingleField_postProcess')) {
                $hookObj->getSingleField_postProcess($table, $field, $row, $out, $PA, $this);
            }
        }
        // Return value (string or array)
        return $out;
    }

    /**
     * Creates a checkbox list (renderMode = "checkbox")
     * (Render function for getSingleField_typeSelect())
     *
     * @param string $table See getSingleField_typeSelect()
     * @param string $field See getSingleField_typeSelect()
     * @param array $row See getSingleField_typeSelect()
     * @param array $PA See getSingleField_typeSelect()
     * @param array $config (Redundant) content of $PA['fieldConf']['config'] (for convenience)
     * @param array $selItems Items available for selection
     * @param string $nMV_label Label for no-matching-value
     * @return string The HTML code for the item
     * @see getSingleField_typeSelect()
     * @todo Define visibility
     */
    public function getSingleField_typeSelect_checkbox($table, $field, $row, &$PA, $config, $selItems, $nMV_label) {
        if (empty($selItems)) {
            return '';
        }
        // Get values in an array (and make unique, which is fine because there can be no duplicates anyway):
        $itemArray = array_flip($this->extractValuesOnlyFromValueLabelList($PA['itemFormElValue']));
        $item = '';
        $disabled = '';
        if ($this->renderReadonly || $config['readOnly']) {
            $disabled = ' disabled="disabled"';
        }
        // Traverse the Array of selector box items:
        $tRows = array();
        $c = 0;
        $setAll = array();
        $unSetAll = array();
        $restoreCmd = array();
        $sOnChange = '';

        $sOnChange = $disabled ? 'return false' : implode('', $PA['fieldChangeFunc']);
        // Used to accumulate the JS needed to restore the original selection.
        foreach ($selItems as $p) {
            // Non-selectable element:
            if ($p[1] === '--div--') {
                $selIcon = '';
                if (isset($p[2]) && $p[2] != 'empty-emtpy') {
                    $selIcon = $this->getIconHtml($p[2]);
                }
                $tRows[] = '
                    <tr class="c-header">
                        <td colspan="3">' . $selIcon . htmlspecialchars($p[0]) . '</td>
                    </tr>';
            } else {
                // Selected or not by default:
                $sM = '';
                if (isset($itemArray[$p[1]])) {
                    $sM = ' checked="checked"';
                    unset($itemArray[$p[1]]);
                }
                // Icon:
                if ($p[2]) {
                    $selIcon = $p[2];
                } else {
                    $selIcon = IconUtility::getSpriteIcon('empty-empty');
                }
                // Compile row:
                $rowId = uniqid('select_checkbox_row_');
                $onClickCell = $this->elName(($PA['itemFormElName'] . '[' . $c . ']')) . '.checked=!' . $this->elName(($PA['itemFormElName'] . '[' . $c . ']')) . '.checked;';
                $onClick = 'this.attributes.getNamedItem("class").nodeValue = ' . $this->elName(($PA['itemFormElName'] . '[' . $c . ']')) . '.checked ? "c-selectedItem" : "c-unselectedItem";';
                $setAll[] = $this->elName(($PA['itemFormElName'] . '[' . $c . ']')) . '.checked=1;';
                $setAll[] .= '$(\'' . $rowId . '\').removeClassName(\'c-unselectedItem\');$(\'' . $rowId . '\').addClassName(\'c-selectedItem\');';
                $unSetAll[] = $this->elName(($PA['itemFormElName'] . '[' . $c . ']')) . '.checked=0;';
                $unSetAll[] .= '$(\'' . $rowId . '\').removeClassName(\'c-selectedItem\');$(\'' . $rowId . '\').addClassName(\'c-unselectedItem\');';
                $restoreCmd[] = $this->elName(($PA['itemFormElName'] . '[' . $c . ']')) . '.checked=' . ($sM ? 1 : 0) . ';' . '$(\'' . $rowId . '\').removeClassName(\'c-selectedItem\');$(\'' . $rowId . '\').removeClassName(\'c-unselectedItem\');' . '$(\'' . $rowId . '\').addClassName(\'c-' . ($sM ? '' : 'un') . 'selectedItem\');';

                if ($disabled) {
                    $onClick = $onClickCell = 'return false';
                    $setAll = $unSetAll = $restoreCmd = array();
                }
                // Check if some help text is available
                // Since TYPO3 4.5 help text is expected to be an associative array
                // with two key, "title" and "description"
                // For the sake of backwards compatibility, we test if the help text
                // is a string and use it as a description (this could happen if items
                // are modified with an itemProcFunc)
                $hasHelp = FALSE;
                $help = '';
                $helpArray = array();
                if (is_array($p[3]) && count($p[3]) > 0 || !empty($p[3])) {
                    $hasHelp = TRUE;
                    if (is_array($p[3])) {
                        $helpArray = $p[3];
                    } else {
                        $helpArray['description'] = $p[3];
                    }
                }
                $label = htmlspecialchars($p[0], ENT_COMPAT, 'UTF-8', FALSE);
                if ($hasHelp) {
                    $help = BackendUtility::wrapInHelp('', '', '', $helpArray);
                }
                $tRows[] = '
                    <tr id="' . $rowId . '" class="' . ($sM ? 'c-selectedItem' : 'c-unselectedItem')
                    . '" onclick="' . htmlspecialchars($onClick) . '"' . ($disabled ? '' : ' style="cursor: pointer;"') . '>
                        <td class="c-checkbox"><input type="checkbox"' . $this->insertDefStyle('check')
                    . ' name="' . htmlspecialchars(($PA['itemFormElName'] . '[' . $c . ']'))
                    . '" value="' . htmlspecialchars($p[1]) . '"' . $sM . ' onclick="' . htmlspecialchars($sOnChange)
                    . '"' . $PA['onFocus'] . $disabled . ' /></td>
                        <td class="c-labelCell" onclick="' . htmlspecialchars($onClickCell) . '">' . $this->getIconHtml($selIcon) . $label . '</td>
                            <td class="c-descr" onclick="' . htmlspecialchars($onClickCell) . '">' . (empty($help) ? '' : $help) . '</td>
                    </tr>';
                $c++;
            }
        }

        // Remaining values (invalid):
        if (count($itemArray) && !$PA['fieldTSConfig']['disableNoMatchingValueElement'] && !$config['disableNoMatchingValueElement']) {
            foreach ($itemArray as $theNoMatchValue => $temp) {
                // Compile <checkboxes> tag:
                array_unshift($tRows, '
						<tr class="c-invalidItem">
							<td class="c-checkbox"><input type="checkbox"' . $this->insertDefStyle('check')
                    . ' name="' . htmlspecialchars(($PA['itemFormElName'] . '[' . $c . ']'))
                    . '" value="' . htmlspecialchars($theNoMatchValue) . '" checked="checked" onclick="' . htmlspecialchars($sOnChange) . '"'
                    . $PA['onFocus'] . $disabled . ' /></td>
							<td class="c-labelCell">' . htmlspecialchars(@sprintf($nMV_label, $theNoMatchValue), ENT_COMPAT, 'UTF-8', FALSE) . '</td><td>&nbsp;</td>
						</tr>');
                $c++;
            }
        }
        // Add an empty hidden field which will send a blank value if all items are unselected.
        $item .= '<input type="hidden" class="select-checkbox" name="' . htmlspecialchars($PA['itemFormElName']) . '" value="" />';
        // Remaining checkboxes will get their set-all link:
        $tableHead = '';
        if (count($setAll)) {
            $tableHead = '<thead>
					<tr class="c-header-checkbox-controls t3-row-header">
						<td class="c-checkbox">
						<input type="checkbox" class="checkbox" onclick="if (checked) {' . htmlspecialchars(implode('', $setAll) . '} else {' . implode('', $unSetAll) . '}') . '">
						</td>
						<td colspan="2">
						</td>
					</tr></thead>';
        }
        // Implode rows in table:
        $item .= '
			<table border="0" cellpadding="0" cellspacing="0" class="typo3-TCEforms-select-checkbox">' . $tableHead . '<tbody>' . implode('', $tRows) . '</tbody>
			</table>
			';
        // Add revert icon
        if (!empty($restoreCmd)) {
            $item .= '<a href="#" onclick="' . implode('', $restoreCmd) . ' return false;' . '">'
                . IconUtility::getSpriteIcon('actions-edit-undo', array('title' => htmlspecialchars($this->getLL('l_revertSelection')))) . '</a>';
        }
        return $item;
    }
}